#ifndef BARVINOK_POLYLIB_H
#define BARVINOK_POLYLIB_H

#include <gmp.h>

#if defined(__cplusplus)
extern "C" {
#endif

#include <polylib/polylibgmp.h>

#undef divide
#undef value_compare

#if defined(__cplusplus)
}
#endif

#endif
